/*******************************************************************************
 * Program..: cistern.ino Monitor rain barrel and workroom temperature
 * Author...: Jan-marie Newton
 * Date.....: 2020-10-10
 * History..: 
 * Notes....: version 2.9
 * problem with ethernet shield https://forum.arduino.cc/index.php?topic=372779.0
 *                 https://forum.arduino.cc/index.php?topic=351477.15
 *******************************************************************************/
#include <OneWire.h>
#include <DallasTemperature.h>
#include <SPI.h>         // needed for Arduino versions later than 0018
#include <Ethernet.h>
#include <EthernetUdp.h> // UDP library from: bjoern@cs.stanford.edu 12/30/2008
#include <TimedAction.h>

// Constants - set relay pin numbers:
//------------------------------------------------------------------------------
#define ONE_WIRE_BUS 3 // digital pin we're connected to
#define RELAY 4        // 1st relay and must be the Fountain relay
#define LEVEL_PIN A0   // analog pin to float switchs
#define PUMP_ON   A3   // catch voltage spike when pump turns on


// Setup a oneWire instance to communicate with any OneWire devices  
//------------------------------------------------------------------------------
// (not just Maxim/Dallas temperature ICs) 
OneWire oneWire(ONE_WIRE_BUS);       //  temp sensor DS18B20
DallasTemperature sensors(&oneWire); // Pass our oneWire reference to Dallas Temperature


// constants won't change. They're used here to
//------------------------------------------------------------------------------
const char* cmds[]       = {"Sleep", "Wake", "Reset", "Adj"};
//const int FLOAT_LEVEL[8] = {1023, 668, 624, 568, 494, 393, 242, 0};
//const int BUTTONS_TOTAL  = 8;

// Initialize Variables
//------------------------------------------------------------------------------
EthernetUDP Udp;         // An EthernetUDP instance to let us send and receive packets over UDP
IPAddress ip(192, 168, 1, 125);
byte mac[]             = {0xDE, 0xAD, 0xBE, 0xEF, 0x1E, 0x1C}; // Enter a MAC address and IP address
unsigned int localPort = 8888;      // local port to listen on
unsigned int debug     = 1;
unsigned int pin       = 0;
int opt       = 0;
int lvl       = 0;
int res       = 0;
int diff      = 0;

// Floats for ADC voltage & Input voltage
float adc_voltage = 0.0;
int in_voltage    = 0.0;
const float voltDivider = (30000.0 + 7500.0) / 7500.0; // resistor values in divider (in ohms) on the
const float ref_voltage = 5.0;     // Float for Reference Voltage


// buffers for receiving and sending data
char packetBuffer[UDP_TX_PACKET_MAX_SIZE];               // buffer to hold incoming packet
char replyBuffer[]     = "0000\t0000\tS\t0000\t5\t00\t00000"; // a string to send back
char allRelays[]       = "0000";
char mode              = 'W';      // S = sleep, W = Wake
float t                = 32;
float o                = 32;
float adj              = 2.1;
unsigned int relays    = sizeof(allRelays) - 1;

// Store Status of temperature and relays
//------------------------------------------------------------------------------
void getStatus() {
   sensors.requestTemperatures();
   t  = sensors.getTempFByIndex(0); // read workroom Temperature as fahrenheit
   o  = sensors.getTempFByIndex(1); // read wine celler Temperature as fahrenheit
   t += adj;

   //  read the bank of relays 
   for(int i=RELAY; i < relays+RELAY; i++) {
      allRelays[i-RELAY] = (digitalRead(i) + '0');
   }

   //  read cistern level
   getWaterLevel();
   /*res  = analogRead(LEVEL_PIN);
   diff = 9999;
   for(int i=0; i < BUTTONS_TOTAL; i++) {
      if (abs(FLOAT_LEVEL[i] - res) < diff) {
         diff = abs(FLOAT_LEVEL[i] - res);
         lvl  = i;
      }
   }*/

   if (debug) {
      Serial.print("inside: ");
      Serial.print(t, 2);    //print the temperature
      Serial.print(" outside: ");
      Serial.print(o, 2);    //print the temperature
      Serial.print(" Mode:");
      Serial.print(mode);
      Serial.print(" Relays:");
      Serial.print(allRelays);
      Serial.print(" Level:");
      Serial.print(lvl);
      Serial.print(" Raw:");
      Serial.print(res);
      Serial.print(" Voltage:");
      Serial.print(in_voltage);
      Serial.print(" adj: ");
      Serial.println(adj,1);
   }

   String rec  = String(round(t * 10)) + "\t"; // inside temperature to 1/10 of degree
   rec += String(round(o * 10)) + "\t";        // outside temperature to 1/10 of degree
   rec += mode; rec += "\t";                   // thermostat mode
   rec += allRelays; rec += "\t";              // current relay dispostion
   rec += lvl; rec += "\t";                    // cistern water level
   rec += in_voltage; rec += "\t";             // voltage to pump
   rec += res;

   int str_len = rec.length() + 1;
   rec.toCharArray(replyBuffer, str_len);
   //Serial.println(replyBuffer);
}
//------------------------------------------------------------------------------

// Set actions to perform on separate timed intervals
//------------------------------------------------------------------------------
/*http://playground.arduino.cc/Code/TimedAction
  http://wiring.uniandes.edu.co/source/trunk/wiring/firmware/libraries/TimedAction
  multi-task ie create a timers that will fire repeatedly every x ms */
TimedAction statusThread = TimedAction(5000, getStatus);
//--------------------------------------------------------------------------

// Setup Initialization
//------------------------------------------------------------------------------
void setup() {

   // start the Ethernet and UDP:
   Ethernet.begin(mac, ip);
   Udp.begin(localPort);

   // initialize the relay pins as an output:
   for(int i=RELAY; i < relays+RELAY; i++) {
      pinMode(i, OUTPUT);
      digitalWrite(i,LOW);
   }
   pinMode(LEVEL_PIN, INPUT);
   pinMode(PUMP_ON, INPUT);

   if (debug) {
     Serial.begin(115200);
     Serial.println("Setup complete");
   }
}
//------------------------------------------------------------------------------

// Main Loop
//------------------------------------------------------------------------------
void loop(){
   // Determine voltage at from PUMP_ON input
   adc_voltage = (analogRead(PUMP_ON) * ref_voltage) / 1024.0;
   // Calculate voltage at divider input
   in_voltage  = (int)round(adc_voltage * voltDivider);

   // get status
   statusThread.check();

   // if there's data available, read a packet
   int packetSize = Udp.parsePacket();
   if (packetSize) {
      Udp.read(packetBuffer, UDP_TX_PACKET_MAX_SIZE); // read the packet into packetBuffer

       /*  Serial.print("Received packet of size ");
           Serial.println(packetSize);
           Serial.print("From ");
           IPAddress remote = Udp.remoteIP();
           for (int i = 0; i < 4; i++) {
              Serial.print(remote[i], DEC);
              if (i < 3) Serial.print(".");
           }
           Serial.print(", port ");
           Serial.println(Udp.remotePort());
           Serial.print("Contents: ");
           Serial.println(packetBuffer); */

      for (int i=0; i < 4; i++){ obeyCommmad(cmds[i]); }
      if (packetBuffer[0] == 'R' && packetBuffer[1] >= '1' && packetBuffer[1] <= (relays + '0')) {
         pin = packetBuffer[1] - 49 + RELAY; // convert to int and sub 1 and add first relay pin
         opt = packetBuffer[2] - '0';
         if (pin == RELAY && mode == 'S' && opt != 0)
            opt = 0;  // cant turn on fountain when in sleep mode
         /* if (debug) {
            Serial.print("set relay opt: ");
            Serial.print(opt);
            Serial.print("  pin: ");
            Serial.println(pin);
         }  */
         switch (opt) {
         case 2:
            opt = 1 - digitalRead(pin);
         case 0:
         case 1:
            digitalWrite(pin, opt);
         }
         getStatus();
      }

      // send a reply to the IP address and port that sent us the packet we received
      Udp.beginPacket(Udp.remoteIP(), Udp.remotePort());
      Udp.write(replyBuffer);
      Udp.endPacket();
   }
}
//------------------------------------------------------------------------------

// Process incoming command
//------------------------------------------------------------------------------
void obeyCommmad(const char* cmd) {
   char *ret;
   ret = strstr(packetBuffer, cmd);
   if (ret == NULL) {
      return;
   }

   if (packetBuffer[0] != cmd[0] || packetBuffer[1] != cmd[1]) return;

   //Serial.print("set command: "); Serial.println(cmd);

   if (cmd == "Sleep") {
      mode = 'S';
      digitalWrite(RELAY, LOW);

   } else if (cmd == "Wake")
      mode = 'W';

   else if (cmd == "Reset")   // initialize the relay pins as an output:
      for(int i=RELAY; i < relays+RELAY; i++)
         digitalWrite(i,LOW);

   else if (cmd[0] == 'A' && cmd[1] == 'd') {
      int a = atoi( &packetBuffer[3] );
      adj   = a / 10.0;
      if (adj > 20) adj = 20;
   }

   getStatus();
}
//------------------------------------------------------------------------------

/* read resistance from tank resistor matrix
 * 1.62m   1.62m    54
 * 368k     300k   252
 * 200k     120k   455
 * 130k      62k   634
 *  50k      28k   803
 *  20k      12k   918
 * 100       99   1023
*------------------------------------------------------------------------------*/
void getWaterLevel() {
   res  = analogRead(LEVEL_PIN);
   if (res < 10)
      lvl = 0;
   else if (res < 100)
      lvl = 1;
   else if (res < 300)
      lvl = 2;
   else if (res < 500)
      lvl = 3;
   else if (res < 700)
      lvl = 4;
   else if (res < 850)
      lvl = 5;
   else if (res < 950)
      lvl = 6;
   else
      lvl = 7;
}
//------------------------------------------------------------------------------
